/*-
 * Copyright (c) 2005 Roman Petrov <melkor@pikenet.ru>
 *                    Aleksandr A Babaylov <.@babolo.ru>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2005 Roman Petrov <melkor@pikenet.ru>, @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: tcpttld.c,v 1.3 2005/05/17 14:05:37 babolo Exp $";
#endif /* not lint */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sysctl.h>
#include <net/if_dl.h>
#include <net/route.h>
#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <signal.h>
#include <paths.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>

#define ADJUST_CHECKSUM(acc, cksum) { \
        acc += cksum; \
        if (acc < 0) { \
                acc = -acc; \
                acc = (acc >> 16) + (acc & 0xffff); \
                acc += acc >> 16; \
                cksum = (u_short) ~acc; \
        } else { \
                acc = (acc >> 16) + (acc & 0xffff); \
                acc += acc >> 16; \
                cksum = (u_short) acc; \
        } \
}

int
main(int argc, char* argv[]) {
    int ch, s, diff, ttl = 1, debug = 0;
    char pkt[IP_MAXPACKET];
    FILE *pid = NULL;
    fd_set fdset;
    struct sockaddr_in sin;
    ssize_t pktlen, hlen;
    u_short port;
    socklen_t sinlen;
    struct ip *pip = (struct ip *)pkt;

    openlog(getprogname(), LOG_CONS | LOG_NDELAY | LOG_PID, LOG_USER);
    port = htons(9999);                                                       /* Default Port is 9999 */
    while ((ch = getopt(argc, argv, "dp:P:t:")) != EOF)
    switch (ch) {
    case 'd':
        debug++;
        break;
    case 't':
        ttl = atoi(optarg);
        if  (ttl < 1 || ttl > 255){
            fprintf(stderr, "Error: TTL %d, must be 0 < TTL < 256\n", ttl);
            exit(EX_USAGE);
        }
        break;
    case 'p':
        port = htons(atoi(optarg));
        if  (atoi(optarg) < 1 || atoi(optarg) > 65535 || port == 0) {
            fprintf(stderr, "Port %d, must be 0 < port < 65536\n", port);
            exit(EX_USAGE);
        }
        break;
    case 'P':
        if  ((pid = fopen(optarg, "w")) == NULL) {
            err(EX_CANTCREAT, "Can't create PID file %s", optarg);
        }
        break;
    default:
        exit(EX_USAGE);
    }
    argc -=optind;
    argv +=optind;

    s = socket(PF_INET, SOCK_RAW, IPPROTO_DIVERT);
    if  (s < 0 || s >= FD_SETSIZE) {
        err(EX_OSERR, "can't create divert socket");
    }
    bzero(&sin, sizeof(sin));
    sin.sin_family = PF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = port;
    if  (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
        err(EX_UNAVAILABLE, "can't bind divert socket");
    }
    if  (!debug && daemon(debug, debug) < 0) {
        err(EX_OSERR, "can't become a daemon");
    }
    if  (pid) {
        fprintf(pid, "%d", getpid());
        fclose(pid);
    }
    if  (debug) fprintf(stderr, "pid=%d socket=%d\n", getpid(), s);
    for (;;) {
        FD_ZERO(&fdset);
        FD_SET(s, &fdset);
        if  (select(s + 1, &fdset, (fd_set *)NULL, (fd_set *)NULL, (struct timeval *)NULL) < 0) {
            syslog(LOG_ERR, "select on divert socket: %m");
            if  (debug) warn("select on divert socket");
            continue;
        }
        if  (FD_ISSET(s, &fdset)) {
            sinlen = sizeof(sin);
            pktlen = recvfrom(s, pkt, sizeof(pkt), 0, (struct sockaddr *)&sin, &sinlen);
            if  (pktlen < 0) {
                if  (errno != EINTR) {
                    syslog(LOG_ERR, "read from divert socket: %m");
                    if  (debug) warn("read from divert socket");
                }
            } else {
                hlen = pip->ip_hl << 2;
                if  (debug) fprintf(stderr, "pktlen=%d, hlen=%d\n", pktlen, hlen);
                if  (  ntohs(pip->ip_len) == pktlen
                    && hlen <= pktlen
                    && pktlen - hlen >= sizeof(struct tcphdr)
                    && ttl < pip->ip_ttl
                    ) {
                    if  (debug) fprintf(stderr, "ttl=%d\n", pip->ip_ttl);
                    diff = pip->ip_ttl - ttl;
                    pip->ip_ttl = ttl;
                    ADJUST_CHECKSUM(diff, pip->ip_sum);
                }
                if  (sendto(s, pkt, pktlen, 0, (struct sockaddr *)&sin, sinlen) < 0) {
                    syslog(LOG_ERR, "write to divert socket: %m");
                    if  (debug) warn("write to divert socket");
}   }   }   }   }
